@@ -2,15 +2,20 @@ package k8shandler
 
 import (
 	"bytes"
+	"context"
 	"fmt"
 	"reflect"
+	"strings"
 
+	"github.com/openshift/cluster-logging-operator/pkg/constants"
 	"github.com/openshift/cluster-logging-operator/pkg/utils"
 	"github.com/sirupsen/logrus"
 	"k8s.io/apimachinery/pkg/api/errors"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/intstr"
 	"k8s.io/client-go/util/retry"
 
+	configv1 "github.com/openshift/api/config/v1"
 	logging "github.com/openshift/cluster-logging-operator/pkg/apis/logging/v1"
 	apps "k8s.io/api/apps/v1"
 	v1 "k8s.io/api/core/v1"
@@ -30,8 +35,11 @@ var (
 )
 
 // CreateOrUpdateVisualization reconciles visualization component for cluster logging
-func (clusterRequest *ClusterLoggingRequest) CreateOrUpdateVisualization() (err error) {
-
+func (clusterRequest *ClusterLoggingRequest) CreateOrUpdateVisualization(proxyConfig *configv1.Proxy) (err error) {
+	if clusterRequest.cluster.Spec.Visualization == nil || clusterRequest.cluster.Spec.Visualization.Type == "" {
+		clusterRequest.removeKibana()
+		return nil
+	}
 	if clusterRequest.cluster.Spec.Visualization.Type == logging.VisualizationTypeKibana {
 
 		if err = clusterRequest.CreateOrUpdateServiceAccount(kibanaServiceAccountName, &kibanaServiceAccountAnnotations); err != nil {
@@ -54,11 +62,15 @@ func (clusterRequest *ClusterLoggingRequest) CreateOrUpdateVisualization() (err
 			return
 		}
 
+		if err = clusterRequest.createOrUpdateKibanaConsoleExternalLogLink(); err != nil {
+			return
+		}
+
 		if err = clusterRequest.RemoveOAuthClient("kibana-proxy"); err != nil {
 			return
 		}
 
-		if err = clusterRequest.createOrUpdateKibanaDeployment(); err != nil {
+		if err = clusterRequest.createOrUpdateKibanaDeployment(proxyConfig); err != nil {
 			return
 		}
 
@@ -71,26 +83,75 @@ func (clusterRequest *ClusterLoggingRequest) CreateOrUpdateVisualization() (err
 
 		printUpdateMessage := true
 		retryErr := retry.RetryOnConflict(retry.DefaultRetry, func() error {
-			if !reflect.DeepEqual(kibanaStatus, cluster.Status.Visualization.KibanaStatus) {
+			if !compareKibanaStatus(kibanaStatus, cluster.Status.Visualization.KibanaStatus) {
 				if printUpdateMessage {
-					logrus.Infof("Updating status of Kibana for %q", cluster.Name)
+					logrus.Infof("Updating status of Kibana")
 					printUpdateMessage = false
 				}
 				cluster.Status.Visualization.KibanaStatus = kibanaStatus
-				return clusterRequest.Update(cluster)
+				return clusterRequest.UpdateStatus(cluster)
 			}
 			return nil
 		})
 		if retryErr != nil {
 			return fmt.Errorf("Failed to update Kibana status for %q: %v", cluster.Name, retryErr)
 		}
-	} else {
-		clusterRequest.removeKibana()
 	}
 
 	return nil
 }
 
+func compareKibanaStatus(lhs, rhs []logging.KibanaStatus) bool {
+	// there should only ever be a single kibana status object
+	if len(lhs) != len(rhs) {
+		return false
+	}
+
+	if len(lhs) > 0 {
+		for index, _ := range lhs {
+			if lhs[index].Deployment != rhs[index].Deployment {
+				return false
+			}
+
+			if lhs[index].Replicas != rhs[index].Replicas {
+				return false
+			}
+
+			if len(lhs[index].ReplicaSets) != len(rhs[index].ReplicaSets) {
+				return false
+			}
+
+			if len(lhs[index].ReplicaSets) > 0 {
+				if !reflect.DeepEqual(lhs[index].ReplicaSets, rhs[index].ReplicaSets) {
+					return false
+				}
+			}
+
+			if len(lhs[index].Pods) != len(rhs[index].Pods) {
+				return false
+			}
+
+			if len(lhs[index].Pods) > 0 {
+				if !reflect.DeepEqual(lhs[index].Pods, rhs[index].Pods) {
+					return false
+				}
+			}
+
+			if len(lhs[index].Conditions) != len(rhs[index].Conditions) {
+				return false
+			}
+
+			if len(lhs[index].Conditions) > 0 {
+				if !reflect.DeepEqual(lhs[index].Conditions, rhs[index].Conditions) {
+					return false
+				}
+			}
+		}
+	}
+
+	return true
+}
+
 func (clusterRequest *ClusterLoggingRequest) removeKibana() (err error) {
 	if clusterRequest.isManaged() {
 		name := "kibana"
@@ -123,6 +184,10 @@ func (clusterRequest *ClusterLoggingRequest) removeKibana() (err error) {
 			return
 		}
 
+		if err = clusterRequest.RemoveConfigMap(constants.KibanaTrustedCAName); err != nil {
+			return
+		}
+
 		if err = clusterRequest.RemoveService(name); err != nil {
 			return
 		}
@@ -130,14 +195,37 @@ func (clusterRequest *ClusterLoggingRequest) removeKibana() (err error) {
 		if err = clusterRequest.RemoveServiceAccount(name); err != nil {
 			return
 		}
+
+		if err = clusterRequest.RemoveConsoleExternalLogLink(name); err != nil {
+			return
+		}
+
 	}
 
 	return nil
 }
 
-func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaDeployment() (err error) {
+func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaDeployment(proxyConfig *configv1.Proxy) (err error) {
 
-	kibanaPodSpec := newKibanaPodSpec(clusterRequest.cluster, "kibana", "elasticsearch")
+	kibanaTrustBundle := &v1.ConfigMap{}
+
+	// Create cluster proxy trusted CA bundle.
+	if proxyConfig != nil {
+		err = clusterRequest.createOrUpdateTrustedCABundleConfigMap(constants.KibanaTrustedCAName)
+		if err != nil {
+			return
+		}
+
+		// kibana-trusted-ca-bundle
+		kibanaTrustBundleName := types.NamespacedName{Name: constants.KibanaTrustedCAName, Namespace: constants.OpenshiftNS}
+		if err := clusterRequest.client.Get(context.TODO(), kibanaTrustBundleName, kibanaTrustBundle); err != nil {
+			if !errors.IsNotFound(err) {
+				return err
+			}
+		}
+	}
+
+	kibanaPodSpec := newKibanaPodSpec(clusterRequest.cluster, "kibana", "elasticsearch.openshift-logging.svc.cluster.local", proxyConfig, kibanaTrustBundle)
 	kibanaDeployment := NewDeployment(
 		"kibana",
 		clusterRequest.cluster.Namespace,
@@ -162,18 +250,107 @@ func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaDeployment() (e
 				if errors.IsNotFound(err) {
 					// the object doesn't exist -- it was likely culled
 					// recreate it on the next time through if necessary
+					logrus.Debugf("Returning nil. The deployment %q was not found even though create previously failed.  Was it culled?", kibanaDeployment.Name)
 					return nil
 				}
 				return fmt.Errorf("Failed to get Kibana deployment: %v", err)
 			}
-			current.Spec = kibanaDeployment.Spec
-			return clusterRequest.Update(current)
+
+			current, different := isDeploymentDifferent(current, kibanaDeployment)
+
+			// Check trustedCA certs have been updated or not by comparing the hash values in annotation.
+			newTrustedCAHashedValue := calcTrustedCAHashValue(kibanaTrustBundle)
+			trustedCAHashedValue, _ := current.Spec.Template.ObjectMeta.Annotations[constants.TrustedCABundleHashName]
+			if trustedCAHashedValue != newTrustedCAHashedValue {
+				different = true
+				if kibanaDeployment.Spec.Template.ObjectMeta.Annotations == nil {
+					kibanaDeployment.Spec.Template.ObjectMeta.Annotations = make(map[string]string)
+				}
+				kibanaDeployment.Spec.Template.ObjectMeta.Annotations[constants.TrustedCABundleHashName] = newTrustedCAHashedValue
+			}
+
+			if different {
+				current.Spec = kibanaDeployment.Spec
+				return clusterRequest.Update(current)
+			}
+			return nil
 		})
 	}
 
 	return nil
 }
 
+func isDeploymentDifferent(current *apps.Deployment, desired *apps.Deployment) (*apps.Deployment, bool) {
+
+	different := false
+
+	// is this needed?
+	if !utils.AreMapsSame(current.Spec.Template.Spec.NodeSelector, desired.Spec.Template.Spec.NodeSelector) {
+		logrus.Debugf("Visualization nodeSelector change found, updating '%s'", current.Name)
+		current.Spec.Template.Spec.NodeSelector = desired.Spec.Template.Spec.NodeSelector
+		different = true
+	}
+
+	// is this needed?
+	if !utils.AreTolerationsSame(current.Spec.Template.Spec.Tolerations, desired.Spec.Template.Spec.Tolerations) {
+		logrus.Debugf("Visualization tolerations change found, updating '%s'", current.Name)
+		current.Spec.Template.Spec.Tolerations = desired.Spec.Template.Spec.Tolerations
+		different = true
+	}
+
+	if isDeploymentImageDifference(current, desired) {
+		logrus.Debugf("Visualization image change found, updating %q", current.Name)
+		current = updateCurrentDeploymentImages(current, desired)
+		different = true
+	}
+
+	if utils.AreResourcesDifferent(current, desired) {
+		logrus.Debugf("Visualization resource(s) change found, updating %q", current.Name)
+		different = true
+	}
+
+	if !utils.EnvValueEqual(current.Spec.Template.Spec.Containers[0].Env, desired.Spec.Template.Spec.Containers[0].Env) {
+		current.Spec.Template.Spec.Containers[0].Env = desired.Spec.Template.Spec.Containers[0].Env
+		different = true
+	}
+
+	return current, different
+}
+
+func isDeploymentImageDifference(current *apps.Deployment, desired *apps.Deployment) bool {
+
+	for _, curr := range current.Spec.Template.Spec.Containers {
+		for _, des := range desired.Spec.Template.Spec.Containers {
+			// Only compare the images of containers with the same name
+			if curr.Name == des.Name {
+				if curr.Image != des.Image {
+					return true
+				}
+			}
+		}
+	}
+
+	return false
+}
+
+func updateCurrentDeploymentImages(current *apps.Deployment, desired *apps.Deployment) *apps.Deployment {
+
+	containers := current.Spec.Template.Spec.Containers
+
+	for index, curr := range current.Spec.Template.Spec.Containers {
+		for _, des := range desired.Spec.Template.Spec.Containers {
+			// Only compare the images of containers with the same name
+			if curr.Name == des.Name {
+				if curr.Image != des.Image {
+					containers[index].Image = des.Image
+				}
+			}
+		}
+	}
+
+	return current
+}
+
 func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaRoute() error {
 
 	cluster := clusterRequest.cluster
@@ -247,6 +424,43 @@ func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaRoute() error {
 	return nil
 }
 
+func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaConsoleExternalLogLink() (err error) {
+	cluster := clusterRequest.cluster
+
+	kibanaURL, err := clusterRequest.GetRouteURL("kibana")
+	if err != nil {
+		return err
+	}
+
+	consoleExternalLogLink := NewConsoleExternalLogLink(
+		"kibana",
+		cluster.Namespace,
+		"Show in Kibana",
+		strings.Join([]string{kibanaURL,
+			"/app/kibana#/discover?_g=(time:(from:now-1w,mode:relative,to:now))&_a=(columns:!(kubernetes.container_name,message),query:(query_string:(analyze_wildcard:!t,query:'",
+			strings.Join([]string{
+				"kubernetes.pod_name:\"${resourceName}\"",
+				"kubernetes.namespace_name:\"${resourceNamespace}\"",
+				"kubernetes.container_name.raw:\"${containerName}\"",
+			}, " AND "),
+			"')),sort:!('@timestamp',desc))"},
+			""),
+	)
+
+	utils.AddOwnerRefToObject(consoleExternalLogLink, utils.AsOwner(cluster))
+
+	// In case the object already exists we delete it first
+	if err = clusterRequest.RemoveConsoleExternalLogLink("kibana"); err != nil {
+		return
+	}
+
+	err = clusterRequest.Create(consoleExternalLogLink)
+	if err != nil && !errors.IsAlreadyExists(err) {
+		return fmt.Errorf("Failure creating Kibana console external log link for %q: %v", cluster.Name, err)
+	}
+	return nil
+}
+
 func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaService() error {
 
 	kibanaService := NewService(
@@ -307,9 +521,12 @@ func (clusterRequest *ClusterLoggingRequest) createOrUpdateKibanaSecret() error
 	return nil
 }
 
-func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasticsearchName string) v1.PodSpec {
-
-	var kibanaResources = cluster.Spec.Visualization.KibanaSpec.Resources
+func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasticsearchName string, proxyConfig *configv1.Proxy, trustedCABundleCM *v1.ConfigMap) v1.PodSpec {
+	visSpec := logging.VisualizationSpec{}
+	if cluster.Spec.Visualization != nil {
+		visSpec = *cluster.Spec.Visualization
+	}
+	var kibanaResources = visSpec.Resources
 	if kibanaResources == nil {
 		kibanaResources = &v1.ResourceRequirements{
 			Limits: v1.ResourceList{v1.ResourceMemory: defaultKibanaMemory},
@@ -359,7 +576,7 @@ func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasti
 		InitialDelaySeconds: 5, TimeoutSeconds: 4, PeriodSeconds: 5,
 	}
 
-	var kibanaProxyResources = cluster.Spec.Visualization.ProxySpec.Resources
+	var kibanaProxyResources = visSpec.ProxySpec.Resources
 	if kibanaProxyResources == nil {
 		kibanaProxyResources = &v1.ResourceRequirements{
 			Limits: v1.ResourceList{v1.ResourceMemory: defaultKibanaProxyMemory},
@@ -388,7 +605,6 @@ func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasti
 		"--tls-cert=/secret/server-cert",
 		"-tls-key=/secret/server-key",
 		"-pass-access-token",
-		"-skip-provider-button",
 	}
 
 	kibanaProxyContainer.Env = []v1.EnvVar{
@@ -403,6 +619,9 @@ func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasti
 		},
 	}
 
+	proxyEnv := utils.SetProxyEnvVars(proxyConfig)
+	kibanaProxyContainer.Env = append(kibanaProxyContainer.Env, proxyEnv...)
+
 	kibanaProxyContainer.Ports = []v1.ContainerPort{
 		{Name: "oaproxy", ContainerPort: 3000},
 	}
@@ -411,6 +630,18 @@ func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasti
 		{Name: "kibana-proxy", ReadOnly: true, MountPath: "/secret"},
 	}
 
+	addTrustedCAVolume := false
+	// If trusted CA bundle ConfigMap exists and its hash value is non-zero, mount the bundle.
+	if trustedCABundleCM != nil && hasTrustedCABundle(trustedCABundleCM) {
+		addTrustedCAVolume = true
+		kibanaProxyContainer.VolumeMounts = append(kibanaProxyContainer.VolumeMounts,
+			v1.VolumeMount{
+				Name:      constants.KibanaTrustedCAName,
+				ReadOnly:  true,
+				MountPath: constants.TrustedCABundleMountDir,
+			})
+	}
+
 	kibanaPodSpec := NewPodSpec(
 		"kibana",
 		[]v1.Container{kibanaContainer, kibanaProxyContainer},
@@ -428,10 +659,32 @@ func newKibanaPodSpec(cluster *logging.ClusterLogging, kibanaName string, elasti
 			},
 			},
 		},
-		cluster.Spec.Visualization.KibanaSpec.NodeSelector,
-		cluster.Spec.Visualization.KibanaSpec.Tolerations,
+		visSpec.NodeSelector,
+		visSpec.Tolerations,
 	)
 
+	if addTrustedCAVolume {
+		optional := true
+		kibanaPodSpec.Volumes = append(kibanaPodSpec.Volumes,
+			v1.Volume{
+				Name: constants.KibanaTrustedCAName,
+				VolumeSource: v1.VolumeSource{
+					ConfigMap: &v1.ConfigMapVolumeSource{
+						LocalObjectReference: v1.LocalObjectReference{
+							Name: constants.KibanaTrustedCAName,
+						},
+						Optional: &optional,
+						Items: []v1.KeyToPath{
+							{
+								Key:  constants.TrustedCABundleKey,
+								Path: constants.TrustedCABundleMountFile,
+							},
+						},
+					},
+				},
+			})
+	}
+
 	kibanaPodSpec.Affinity = &v1.Affinity{
 		PodAntiAffinity: &v1.PodAntiAffinity{
 			PreferredDuringSchedulingIgnoredDuringExecution: []v1.WeightedPodAffinityTerm{
